import tensorflow as tf
import sys


def float_feature(list_of_floats):
    return tf.train.Feature(float_list=tf.train.FloatList(value=list_of_floats))


def int64_feature(list_of_ints):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=list_of_ints))


def bytestring_feature(list_of_bytestrings):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=list_of_bytestrings))


def print_one_line(*args):
    tf.print("\033[K", end="")
    tf.print("\r", *args, sep="", end=" ", output_stream=sys.stdout)



import re
from typing import Dict

import yaml

from vistec_ser.augmentations import Augmentation


def load_yaml(path: str) -> Dict[str, object]:
    # Fix yaml numbers https://stackoverflow.com/a/30462009/11037553
    loader = yaml.SafeLoader
    loader.add_implicit_resolver(
        u'tag:yaml.org,2002:float',
        re.compile(u'''^(?:
         [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$''', re.X),
        list(u'-+0123456789.'))
    with open(path, "r", encoding="utf-8") as file:
        return yaml.load(file, Loader=loader)


class FeatureConfig:
    def __init__(self, config: Dict[str, any]):
        self.sample_rate = config.get('sample_rate', 16000)
        self.frame_length = config.get('frame_length', 50)
        self.frame_shift = config.get('frame_shift', 10)
        self.num_mel_bins = config.get('num_mel_bins', 40)
        self.feature_type = config.get('feature_type', 'fbank')
        self.preemphasis = config.get('preemphasis', 0.97)
        self.scale_normalize = config.get('scale_normalize', False)


class Config:
    def __init__(self, path: str):
        config = load_yaml(path)
        self.augmentations = Augmentation(config=config.get('augmentations', {}))
        self.feature_config = config.get('feature_config', {})


from typing import List

from tensorflow.keras import Model
from tensorflow.keras.layers import Conv1D, BatchNormalization, LSTM, Bidirectional, Dense, Masking
import tensorflow as tf

from .layers.cnn_blocks import CNN1DBlock


def get_cnn(filters: int, kernel_size: int, stride: int = 1, activation: str = 'relu'):
    return tf.keras.Sequential([
        Conv1D(filters=filters, kernel_size=kernel_size, strides=stride, activation=activation),
        BatchNormalization()
    ])


class Wav2Letter(Model):

    def __init__(self, n_classes: int, activation: str = 'relu', **kwargs):
        super().__init__(**kwargs)
        self.activation = activation
        self.conv_layers = [
            get_cnn(250, 48, stride=2, activation=self.activation),
            get_cnn(250, 7, stride=1, activation=self.activation),
            get_cnn(250, 7, stride=1, activation=self.activation),
            get_cnn(250, 7, stride=1, activation=self.activation),
            get_cnn(250, 7, stride=1, activation=self.activation),
            get_cnn(250, 7, stride=1, activation=self.activation),
            get_cnn(250, 7, stride=1, activation=self.activation),
            get_cnn(250, 7, stride=1, activation=self.activation),
            get_cnn(2000, 32, stride=1, activation=self.activation),
            get_cnn(2000, 1, stride=1, activation=self.activation),
            get_cnn(n_classes, 1, stride=1, activation=self.activation),
        ]

    def call(self, x, training=False, **kwargs):
        for cnn in self.conv_layers:
            x = cnn(x, training=training)
        x = tf.reduce_mean(x, axis=1)
        x = tf.nn.softmax(x, axis=-1)
        return x

    def get_config(self):
        super().get_config()


class CNN1DLSTM(Model):
    def __init__(
            self,
            n_classes: int,
            num_channels: List[int] = [64, 64, 128, 128],
            kernel_sizes: List[int] = [4, 3, 3, 3],
            pool_sizes: List[int] = [2, 2, 2, 2],
            bidirectional: bool = True,
            rnn_units: List[int] = [128],
            norm_method: str = 'layer',
            dropout: float = 0.5,
            **kwargs):
        super().__init__(**kwargs)

        self.num_channels = num_channels
        self.kernel_sizes = kernel_sizes
        self.pool_sizes = pool_sizes
        self.bidirectional = bidirectional
        self.rnn_units = rnn_units
        self.n_classes = n_classes
        self.dropout = dropout

        # CNN layers
        assert len(num_channels) == len(kernel_sizes) == len(
            pool_sizes), 'Number of channel / kernel / pool sizes mismatch: {}, {} and {}'.format(len(num_channels),
                                                                                                  len(kernel_sizes),
                                                                                                  len(pool_sizes))
        self.cnn_layers = [CNN1DBlock(
            n_channel=c,
            kernel_size=k,
            pool_size=p,
            norm_method=norm_method
        ) for c, k, p in zip(num_channels, kernel_sizes, pool_sizes)]

        # RNN layer
        self.recurrent_layers = [
            LSTM(unit, return_sequences=True) if i != len(rnn_units)-1 else LSTM(unit)
            for i, unit in enumerate(rnn_units)
        ]
        if bidirectional:
            self.recurrent_layers = [Bidirectional(layer) for layer in self.recurrent_layers]
        # logits layer
        self.logits = Dense(n_classes)

    def get_config(self):
        return {'num_channels': self.num_channels,
                'kernel_sizes': self.kernel_sizes,
                'pool_sizes': self.pool_sizes,
                'bidirectional': self.bidirectional,
                'rnn_units': self.rnn_units,
                'n_classes': self.n_classes,
                'dropout': self.dropout}

    def call(self, x, training=False, **kwargs) -> tf.Tensor:
        # cnn layers
        for cnn_block in self.cnn_layers:
            x = cnn_block(x, training=training)

        # masking
        x = Masking()(x)

        # lstm
        for rnn in self.recurrent_layers:
            x = rnn(x)

        # dense
        if training:
            x = tf.nn.dropout(x, rate=self.dropout)
        x = self.logits(x)
        return tf.nn.softmax(x, axis=-1)


class TestModel(Model):

    def __init__(self, n_classes: int = 4, **kwargs):
        super().__init__(**kwargs)
        self.logits = Dense(n_classes)

    def call(self, x, **kwargs):
        x = tf.reduce_mean(x, axis=1)
        x = self.logits(x)
        x = tf.nn.softmax(x, axis=-1)
        return x

    def get_config(self):
        super().get_config()



from typing import List, Dict, Any

import tensorflow as tf


def load_waveform(
        audio: str,
        bit_depth: int = 16,
        dither: float = 1.) -> tf.Tensor:
    audio_binary = tf.io.read_file(audio)
    waveform, sr = tf.audio.decode_wav(audio_binary)
    # if sr != sample_rate:
    #     raise ValueError(f"Expected sampling rate of {sample_rate} but got {sr}")
    waveform = tf.squeeze(waveform, axis=-1)
    waveform = waveform + tf.random.normal(tf.shape(waveform), stddev=(dither / 2 ** bit_depth))
    return waveform


def pre_emphasis(waveform: tf.Tensor, coeff: float = 0.97) -> tf.Tensor:
    if isinstance(coeff, float) or coeff <= 0.0:
        return waveform
    return tf.concat([waveform[0], waveform[1:] - coeff * waveform[:-1]])


def normalize_feature(feature: tf.Tensor, scaled: bool = False, epsilon: float = 1e-8) -> tf.Tensor:
    mu = tf.reduce_mean(feature, axis=0)
    sum_x2 = tf.reduce_sum(feature ** 2, axis=0) / tf.cast(tf.shape(feature)[0], dtype=tf.float32)
    sigma = sum_x2 - (mu ** 2) + epsilon
    normalized = (feature - mu) / sigma if scaled else feature - mu
    return normalized


def dbscale(spectrogram: tf.Tensor, top_db: int = 80):
    power = tf.math.square(spectrogram)
    log_spec = 10.0 * (tf.math.log(power) / tf.math.log(10.0))
    log_spec = tf.math.maximum(log_spec, tf.math.reduce_max(log_spec) - top_db)
    return log_spec


# ----------------- VTLP ----------------------


def get_scale_factors(freq_dim, sampling_rate, f_hi=4800., alpha=0.9):
    factors = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
    freqs = tf.cast(tf.linspace(0, 1, freq_dim), tf.float32)

    scale = f_hi * tf.minimum(alpha, 1.)
    f_boundary = scale / alpha
    half_sr = sampling_rate / 2

    for i in tf.range(len(freqs)):
        f = sampling_rate * freqs[i]
        if tf.math.less_equal(f, f_boundary):
            factors = factors.write(factors.size(), f * alpha)
        else:
            warp_freq = half_sr - (half_sr - scale) / (half_sr - scale / alpha) * (half_sr - f)
            factors = factors.write(factors.size(), warp_freq)
    return factors.stack()


def apply_vtlp(spec, sampling_rate: int = 16000, f_hi: int = 4800, alpha=0.9):
    time_dim, freq_dim = tf.shape(spec)[0], tf.shape(spec)[1]
    factors = get_scale_factors(freq_dim, sampling_rate, f_hi=f_hi, alpha=alpha)
    factors = factors * tf.cast(freq_dim - 1, tf.float32) / tf.reduce_max(factors)
    new_spec = tf.zeros([freq_dim, time_dim], dtype=tf.float32)

    for i in tf.range(freq_dim):
        if tf.equal(i, 0) or tf.greater_equal(i + 1, freq_dim):
            update = tf.expand_dims(spec[:, i], 0)
            new_spec = tf.tensor_scatter_nd_add(new_spec, [[i]], update)
        else:
            warp_up = factors[i] - tf.floor(factors[i])
            warp_down = 1. - warp_up
            pos = tf.cast(tf.floor(factors[i]), tf.int64)

            new_spec = tf.tensor_scatter_nd_add(new_spec, [[pos]], tf.expand_dims(warp_down * spec[:, i], 0))
            new_spec = tf.tensor_scatter_nd_add(new_spec, [[pos + 1]], tf.expand_dims(warp_up * spec[:, i], 0))
    return tf.transpose(new_spec, perm=(1, 0))


# ----------------- FeatureLoader -----------------


class FeatureLoader:
    def __init__(self, config: Dict[str, object]):
        self.sample_rate = config.get('sample_rate', 16000)
        self.max_length = int(self.sample_rate * (config.get('max_length', None) / 1000))  # in ms

        self.frame_length = int(self.sample_rate * (config.get("frame_length", 25) / 1000))
        self.frame_shift = int(self.sample_rate * (config.get("frame_shift", 10) / 1000))

        self.num_mel_bins = config.get('num_mel_bins', 80)
        self.feature_type = config.get('feature_type', 'fbank')
        self.preemphasis = config.get('preemphasis', 0.97)

        self.vtlp = config.get('vtlp', {})
        if len(self.vtlp) > 0:
            self.alpha_range = tf.range(
                self.vtlp.get("min_alpha", 0.9),
                self.vtlp.get("max_alpha", 1.1),
                self.vtlp.get("alpha_step", 0.1),
            )
        else:
            self.alpha_range = tf.ones([1])
        self.scale_normalize = config.get('scale_normalize', True)

    @property
    def shape(self) -> List[Any]:
        if self.feature_type == 'wave':
            return [self.max_length]
        elif self.feature_type == 'fbank':
            frame_shift = self.frame_shift * 1000. / self.sample_rate
            time_dim = int((self.max_length / self.sample_rate) * (1000. / frame_shift)) if self.max_length else None
            return [time_dim, self.num_mel_bins]
        else:
            frame_shift = self.frame_shift * 1000. / self.sample_rate
            time_dim = int((self.max_length / self.sample_rate) * (1000. / frame_shift)) if self.max_length else None
            return [time_dim, None]

    def sample_alpha(self):
        idxs = tf.range(tf.shape(self.alpha_range)[0])
        ridxs = tf.random.shuffle(idxs)[0]
        sampled_alpha = tf.gather(self.alpha_range, ridxs)
        return sampled_alpha

    def stft(self, waveform: tf.Tensor) -> tf.Tensor:
        spectrogram = tf.abs(
            tf.signal.stft(
                waveform,
                frame_length=self.frame_length,
                frame_step=self.frame_shift,
                window_fn=tf.signal.hann_window,
                pad_end=True
            )
        )
        return spectrogram
        # return apply_vtlp(spectrogram, alpha=self.sample_alpha())

    def make_spectrogram(self, waveform: tf.Tensor) -> tf.Tensor:
        spectrogram = self.stft(waveform)
        return dbscale(spectrogram)

    def make_fbank(self, waveform: tf.Tensor, fmin: float = 0., fmax: float = 8000.) -> tf.Tensor:
        spectrogram = self.stft(waveform)
        mel_fbank = tf.signal.linear_to_mel_weight_matrix(
            num_mel_bins=self.num_mel_bins,
            num_spectrogram_bins=spectrogram.shape[-1],
            sample_rate=self.sample_rate,
            lower_edge_hertz=fmin, upper_edge_hertz=fmax
        )
        fbank = tf.tensordot(spectrogram, mel_fbank, 1)
        return dbscale(fbank)

    def make_mfcc(self, waveform: tf.Tensor) -> tf.Tensor:
        log_mel_spectrogram = self.make_fbank(waveform)
        return tf.signal.mfccs_from_log_mel_spectrograms(log_mel_spectrogram)

    def extract(self, waveform: tf.Tensor) -> tf.Tensor:
        waveform = waveform[:self.max_length] if self.max_length else waveform
        waveform = pre_emphasis(waveform, coeff=self.preemphasis)

        if self.feature_type == "wave":
            return waveform
        elif self.feature_type == "spectrogram":
            features = self.make_spectrogram(waveform)
        elif self.feature_type == "fbank":
            features = self.make_fbank(waveform)
        elif self.feature_type == "mfcc":
            features = self.make_mfcc(waveform)
        else:
            raise KeyError(f"Unsupported feature type `{self.feature_type}`. Only `spectrogram`, `fbank`, and `mfcc` "
                           f"available.")

        return normalize_feature(features, scaled=self.scale_normalize)



from typing import Dict

from .spec_augments import FreqMasking, TimeMasking


AUGMENTATIONS = {
    "freq_masking": FreqMasking,
    "time_masking": TimeMasking,
}


class SpecFlow:
    def __init__(self, flow: list):
        self.flow = flow

    def augment(self, data):
        for f in self.flow:
            data = f.augment(data)
        return data


class Augmentation:
    def __init__(self, config: Dict[str, object] = None):
        if not config:
            config = dict()
        self.vtlp = config.get("wave_augment", {})
        self.feat_augment = self.parse_spec(config.get("feature_augment", {}))

    @staticmethod
    def parse_spec(config: dict):
        augmentations = list()
        for k, v in config.items():
            aug = AUGMENTATIONS.get(k, None)
            if aug is None:
                raise KeyError(f"Augmentation methods `{k}` not available. Only {AUGMENTATIONS.keys()} available.")
            augment = aug(**v) if v is not None else aug()
            augmentations.append(augment)
        return SpecFlow(augmentations)



import tensorflow_io as tfio


# --------------------- FREQ MASKING


class FreqMasking:
    def __init__(self,
                 num_masks: int = 1,
                 mask_factor: float = 27.):
        self.mask_factor = mask_factor
        self.num_masks = num_masks

    def augment(self, data):
        for _ in range(self.num_masks):
            data = tfio.experimental.audio.freq_mask(data, param=self.mask_factor)
        return data


# --------------------- TIME MASKING


class TimeMasking:
    def __init__(self,
                 num_masks: int = 1,
                 mask_factor: float = 40.):
        self.mask_factor = mask_factor
        self.num_masks = num_masks

    def augment(self, data):
        for _ in range(self.num_masks):
            data = tfio.experimental.audio.time_mask(data, param=self.mask_factor)
        return data



from typing import List

from tensorflow.keras.layers import Bidirectional, LSTM, GRU, SimpleRNN, TimeDistributed, Dense, Layer
import tensorflow as tf


class BiPyramidRNN(Layer):
    def __init__(self,
                 units: int,
                 n_layers: int,
                 rnn_type: str = 'lstm'):
        super().__init__()
        if rnn_type == 'lstm':
            self.rnn_stack = [Bidirectional(LSTM(units, return_sequences=True)) for _ in range(n_layers)]
        elif rnn_type == 'gru':
            self.rnn_stack = [Bidirectional(GRU(units, return_sequences=True)) for _ in range(n_layers)]
        elif rnn_type == 'rnn':
            self.rnn_stack = [Bidirectional(SimpleRNN(units, return_sequences=True)) for _ in range(n_layers)]
        else:
            raise NameError('Unrecognized rnn_type: {}'.format(rnn_type))

    def call(self, x, **kwargs):
        i = 0
        for rnn in self.rnn_stack:
            if i == 0:
                x = rnn(x)
            else:
                if tf.shape(x)[1] % 2 == 1:
                    shape = tf.shape(x)
                    pad = tf.zeros([shape[0], 1, shape[-1]])
                    x = tf.concat([x, pad], axis=1)
                x = tf.concat([x[:, 0::2, :], x[:, 1::2, :]], axis=-1)
                x = rnn(x)
            i += 1
        return x


class AttentiveRNN(Layer):
    def __init__(self,
                 units: List[int] = [128],
                 recurrent_type: str = 'gru',
                 bidirectional: bool = True):
        super().__init__()
        if recurrent_type == 'gru':
            self.RNNs = [GRU(unit, return_sequences=True) for unit in units]
        elif recurrent_type == 'lstm':
            self.RNNs = [LSTM(unit, return_sequences=True) for unit in units]
        elif recurrent_type == 'rnn':
            self.RNNs = [SimpleRNN(unit, return_sequences=True) for unit in units]
        else:
            raise NameError('Unknown recurrent type: {}'.format(recurrent_type))

        if bidirectional:
            self.RNNs = [Bidirectional(layer) for layer in self.RNNs]

        self.f_attn = TimeDistributed(Dense(1, activation='tanh'))

    def call(self, x, training=False, return_attn=True, **kwargs):
        for rnn in self.RNNs:
            x = rnn(x)

        attn_score = tf.nn.softmax(self.f_attn(x), axis=1)
        x = tf.reduce_sum(x * attn_score, axis=1)
        if return_attn:
            return x, attn_score
        return x



import setuptools

requirements = [
    "tensorflow>=2.4.0",
    "tensorflow_io>=0.17",
    "numpy>=1.19.2,<1.20.0",
    "soundfile>=0.10.3",
    "PyYAML>=5.3.1",
]

setuptools.setup(
    name="vistec-ser",
    version="0.2a2",
    author="Chompakorn Chaksangchaichot",
    author_email="chompakorn.cc@gmail.com",
    description="Speech Emotion Recognition models and training using Tensorflow 2.x",
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/tann9949/vistec-ser',
    packages=setuptools.find_packages(include=['vistec_ser*']),
    install_requires=requirements,
    classifiers=[
        'Development Status :: 3 - Alpha',  # "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.6'
)



from typing import List, Union
import multiprocessing as mp

from glob import glob
import tensorflow as tf
import numpy as np
import os

from .features import load_waveform, FeatureLoader
from ..augmentations.augmentation import Augmentation
from ..utils import bytestring_feature, print_one_line


AUTOTUNE = tf.data.experimental.AUTOTUNE
TFRECORD_SHARDS = 16
EMOTIONS = ['neutral', 'anger', 'sadness', 'happiness']


def to_tfrecord(path: bytes, audio: bytes, emotion: int) -> tf.train.Example:
    feature = {
        "path": bytestring_feature([path]),
        "audio": bytestring_feature([audio]),
        "emotion": bytestring_feature([emotion])
    }
    return tf.train.Example(features=tf.train.Features(feature=feature))


def write_tfrecord_file(splitted_csv: List[np.ndarray]):
    shard_path, data = splitted_csv
    with tf.io.TFRecordWriter(shard_path, options='ZLIB') as out:
        for audio_path, emotion in data:
            audio = open(audio_path, 'rb').read()
            example = to_tfrecord(
                path=bytes(audio_path, 'utf-8'),
                audio=audio,
                emotion=emotion
            )
            out.write(example.SerializeToString())
            print_one_line("Processed:", audio_path)
    print(f"\nCreated {shard_path}")


class DataLoader:
    def __init__(self,
                 csv_paths: Union[List[str], str],
                 feature_loader: FeatureLoader,
                 augmentations: Augmentation = Augmentation(None),
                 shuffle: bool = True):
        if isinstance(csv_paths, str):
            csv_paths = [csv_paths]
        self.csv_paths = csv_paths
        self.shuffle = shuffle
        self.feature_loader = feature_loader
        self.augmentations = augmentations
        self.steps_per_epoch = None

    def read_csv(self) -> np.ndarray:
        lines = list()
        for csv in self.csv_paths:
            print(f"Reading {csv}...")
            with tf.io.gfile.GFile(csv, 'r') as f:
                lines += f.read().splitlines()[1:]  # skip header
        lines = np.array([line.split(',') for line in lines])  # line i: audio_path, emotion
        if self.shuffle:
            np.random.shuffle(lines)
        return lines

    def get_dataset(self, batch_size) -> Union[tf.data.Dataset, None]:
        data = self.read_csv()
        if len(data) == 0:
            return None
        dataset = tf.data.Dataset.from_tensor_slices(data)
        return self.process_dataset(dataset, batch_size=batch_size)

    @tf.function
    def preprocess(self, data):
        audio_path = data[0]
        emotion = data[1]
        with tf.device('/CPU:0'):
            waveform = load_waveform(audio_path)

            # apply augmentation and feature extraction
            features = self.feature_loader.extract(waveform)
            features = self.augmentations.feat_augment.augment(features)

            label = tf.argmax(tf.stack([tf.equal(emotion, e) for e in EMOTIONS], axis=-1))
            return features, label

    @tf.function
    def parse(self, data):
        return tf.numpy_function(
            self.preprocess,
            inp=[data[0], data[1]],
            Tout=[tf.float32, tf.int64]
        )

    def process_dataset(self,
                        dataset: Union[tf.data.Dataset, tf.data.TFRecordDataset],
                        batch_size: int) -> tf.data.Dataset:
        dataset = dataset.map(self.preprocess, num_parallel_calls=AUTOTUNE)

        if self.shuffle:
            dataset = dataset.shuffle(TFRECORD_SHARDS, reshuffle_each_iteration=True)

        dataset = dataset.padded_batch(
            batch_size=batch_size,
            padded_shapes=(
                tf.TensorShape(self.feature_loader.shape),
                tf.TensorShape([])
            ),
            padding_values=(0., None),
            drop_remainder=False
        )

        dataset = dataset.prefetch(AUTOTUNE)
        self.steps_per_epoch = len(dataset)
        return dataset.repeat(None)


class TFRecordDataset(DataLoader):
    def __init__(self,
                 csv_paths: Union[List[str], str],
                 tfrecords_dir: str,
                 stage: str,  # name of tfrecord (EX: `train_1.tfrecord` for 1st shard of train set tfrecord)
                 feature_loader: FeatureLoader,
                 augmentations: Augmentation = Augmentation(None),
                 shuffle: bool = True):
        super().__init__(self,
                         csv_paths=csv_paths,
                         feature_loader=feature_loader,
                         augmentations=augmentations,
                         shuffle=shuffle)
        self.stage = stage
        self.tfrecords_dir = tfrecords_dir
        if not os.path.exists(self.tfrecords_dir):
            os.makedirs(self.tfrecords_dir)

    def create_tfrecords(self):
        if glob(os.path.join(self.tfrecords_dir, f'{self.stage}*.tfrecord')):
            print("TFRecord files exists.")
            return True

        print(f"Creating {self.stage}.tfrecord...")

        csv = self.read_csv()
        if len(csv) == 0:
            return False

        def get_shard_path(shard_id):
            return os.path.join(self.tfrecords_dir, f'{self.stage}_{shard_id}.tfrecord')

        shards = [get_shard_path(i+1) for i in range(TFRECORD_SHARDS)]
        splitted_entries = np.array_split(csv, TFRECORD_SHARDS)

        with mp.Pool(TFRECORD_SHARDS) as pool:
            pool.map(write_tfrecord_file, zip(shards, splitted_entries))

        return True

    @tf.function
    def parse(self, record):
        feature_description = {
            "path": tf.io.FixedLenFeature([], tf.string),
            "audio": tf.io.FixedLenFeature([], tf.string),
            "emotion": tf.io.FixedLenFeature([], tf.string)
        }
        example = tf.io.parse_single_example(record, feature_description)

        return tf.numpy_function(
            self.preprocess,
            inp=[example["audio"], example["emotion"]],
            Tout=[tf.float32, tf.int64]
        )

    def get_dataset(self, batch_size: int) -> Union[tf.data.Dataset, None]:
        if not self.create_tfrecords:
            return None

        pattern = os.path.join(self.tfrecords_dir, f"{self.stage}*.tfrecord")
        files_ds = tf.data.Dataset.list_files(pattern)
        ignore_order = tf.data.Options()
        ignore_order.experimental_deterministic = False
        files_ds = files_ds.with_options(ignore_order)
        dataset = tf.data.TFRecordDataset(files_ds, compression_type='ZLIB', num_parallel_reads=AUTOTUNE)

        return self.process_dataset(dataset, batch_size)



import tensorflow as tf
from tensorflow.keras.layers import Conv1D, MaxPool1D, Conv2D, MaxPool2D, BatchNormalization, LayerNormalization, Layer


class CNN1DBlock(Layer):
    def __init__(self,
                 n_channel: int,
                 kernel_size: int,
                 pool_size: int,
                 stride: int = 1,
                 padding: str = 'same',
                 norm_method: str = 'layer'):
        super().__init__()
        self.cnn = Conv1D(n_channel, kernel_size, strides=stride, padding=padding)
        self.pool = MaxPool1D(pool_size)
        if norm_method:
            if norm_method == 'layer':
                self.norm = LayerNormalization()
            elif norm_method == 'batch':
                self.norm = BatchNormalization()
            else:
                raise ValueError('Unrecognized normaliziation method: {}'.format(norm_method))
        else:
            self.norm = None

    def call(self, x, training=False):
        x = self.cnn(x)
        if self.norm:
            x = self.norm(x, training=training)
        x = tf.nn.leaky_relu(x, alpha=0.1)
        x = self.pool(x)
        return x
